/**************************************************************************
** TSTTY -- Drive a ptty <--> IP port link
** Copyright 1994 David Schmitt (dschmitt@interaccess.com)
** This software can be freely used for noncommercial purposes.
**
**
** SYNOPSIS:
**	This program opens an IP streams socket to a named host/port #
**	and pipes data to/from a specified ptty device. It assumes that
**	the socket is 8-bit clean, full duplex. No OOB data is handled.
**
**	tstty runs as a daemon and can be killed in the normal manner
**	(SIGINT, SIGQUIT). Fatal diagnostics are sent to stdout.
**
** USE:
**	tstty [hostname] [port #] [ptty #]
**	[hostname]	Hostname to connect to (i.e. ets1.domain.com)
**	[port #]	Port # to open (i.e. 3001 for an ETS8 port 1)
**	[ptty #]	ptty device to claim (i.e. p1 for /dev/ttyp1)
**
** DISCLAIMER:
**	This software is distributed "as-is" without warranties as to it's
**	fitness for any particular use.
**
** HISTORY:
**	Ver		Comment
**	--		-------
**	1		Esta vivo!
**	2		Added command line interface
**	3		Started using select() -- now only 1 process
**	4		Clean up
**
** Changes/enhancements 1995 by Michael Grube <m.grube@heather.heide.de>:
** - made it compile and run under Linux
** - created two different version: tstty and tsttyd. 
** - changed the sizes of sbuf and ptbuf.
** - added response to SIGTERM
** - put in SO_KEEPALIVE so tstty will quit with a SIGPIPE when the
**   connection should be lost.
** - now writes/deletes pid file.
***************************************************************************/
#include <sys/time.h>		/* mg */
#include <sys/types.h>
#include <unistd.h>		/* mg */
#include <errno.h>		/* mg */
#include <sys/socket.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <stdio.h>
#include <signal.h>
#include <netdb.h>

char ourtty[15];

/* File/socket handles */
static int s,pt;	

/* ----------------------------------------------------------------------- */
/* our termination function */
void bye(int sig)
{
	char fn[128];
	
	close(s);
	close(pt);

	strcpy(fn, PIDDIR);
	strcat(fn, "/tstty.");
	strcat(fn, ourtty);
	strcat(fn, ".pid");
	unlink(fn);
    
	exit(0);
}
/* ----------------------------------------------------------------------- */
void write_pidfile()
{
    char fn[128];
    FILE *fp = NULL;
    
    strcpy(fn, PIDDIR);
    strcat(fn, "/tstty.");
    strcat(fn, ourtty);
	strcat(fn, ".pid");
    
    fp = fopen(fn,"w");
    if (fp) {
	fprintf(fp,"%d\n",(int)getpid());
	fclose(fp);
    }
}
/* ----------------------------------------------------------------------- */
void usage(char *p)
{
	if(p) printf("\n%s\n",p);
	printf("Usage: tstty [hostname] [port #] [ptty #]\n");
	printf("[hostname]     Name of remote machine (ie ets1.domain.com)\n");
	printf("[port #]       Port # to connect to (ie 3001)\n");
	printf("[ptty #]       Ptty device to use (ie p1 for ttyp1)\n");
	exit(1);
}
/* ----------------------------------------------------------------------- */
main(int argc,char *argv[])
{
	int on = 1;
	int pp,slen,plen,maxfd;
	char sbuf[4096],ptbuf[4096];
	static char buf[130];
	struct hostent *hp;
	struct sockaddr_in sin;
	fd_set ibits;
	
	/* Do a little sanity checking */
	if(argc < 4) usage(NULL);
	if(sscanf(argv[2],"%d",&pp) != 1) usage("Can't parse port #");
	
	/* Connect to the TS */
	if((hp = gethostbyname(argv[1])) == 0)
	{
		perror("TSTTY: Host name lookup failed ");
		exit(1);
	}
	bzero(&sin,sizeof(sin));
	bcopy(hp->h_addr_list[0],&(sin.sin_addr),hp->h_length);
	sin.sin_family = hp->h_addrtype;

	/* Setup port # */
	sin.sin_port = htons(pp);
	
	if((s = socket(AF_INET,SOCK_STREAM,6)) == -1)
	{
		perror("TSTTY: socket() failed ");
		exit(1);
	}
	
	if(connect(s, (char *)&sin, sizeof(sin)) == -1)
	{
		perror("TSTTY: connect() failed ");
		exit(1);
	}
	
	if (setsockopt(s, SOL_SOCKET, SO_KEEPALIVE, &on, sizeof(on)) < 0) {
		perror("TSTTY: SO_KEEPALIVE on failed ");
		exit(1);
	}
	
	/* Open the designated ptty master */
	sprintf(buf,"/dev/pty%s",argv[3]);
	strcpy(ourtty, argv[3]);
	if((pt = open(buf,O_RDWR)) == -1)
	{
		sprintf(ptbuf,"TSTTY: opening %s ",buf); 
		perror(ptbuf);
		exit(1);
	}

	signal(SIGHUP,  bye);
	signal(SIGPIPE, bye);
	signal(SIGINT,  bye);
	signal(SIGQUIT, bye);
	signal(SIGTERM, bye);

	/* Close stdin, stdout, stderror */
	close(0);
	close(1);
	close(2);

	if(fork() != 0) { 
	    exit(0);
	}
	    
	setpgrp();
	write_pidfile();
	

	/* Update maxfd for select() */
	if(s > pt) maxfd = s + 1;
	else maxfd = pt + 1;

	/* Wait for I/O and process it */
	while(1)
	{
		/* Hoist our flags */
		FD_ZERO(&ibits);
		FD_SET(s,&ibits);
		FD_SET(pt,&ibits);
		
		/* Wait for input */
		if(select(maxfd,&ibits,NULL,NULL,NULL) < 0)
		{
			if(errno == EINTR) continue; /* Should be caught */
			bye(0);
		}

		if((FD_ISSET(s,&ibits)))
		{
			/* Incoming data from TS */
			slen = read(s,sbuf,sizeof(sbuf));
			if(slen) write(pt,sbuf,slen);
		}

		if((FD_ISSET(pt,&ibits)))
		{
			/* Read data from pty */
			plen = read(pt,ptbuf,sizeof(ptbuf));
			if(plen <= 0)
			{
				/* No data? Slave side must've closed! */
				/* Close and reopen the master side */
				close(pt);
				if((pt = open(buf,O_RDWR)) == -1)
				{
					bye(0);
				}
				/* Update maxfd for select() */
				if(s > pt) maxfd = s + 1;
				else maxfd = pt + 1;
			}
			else write(s,ptbuf,plen);
		}
	}
	bye(0);
}
/* END */
/* ----------------------------------------------------------------------- */
